@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtGui module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qquicktextcontrol_p.h"
 #include "qquicktextcontrol_p_p.h"
@@ -72,6 +36,7 @@
 #include <qmetaobject.h>
 
 #include <private/qqmlglobal_p.h>
+#include <private/qquickdeliveryagent_p_p.h>
 
 // ### these should come from QStyleHints
 const int textCursorWidth = 1;
@@ -115,6 +80,8 @@ QQuickTextControlPrivate::QQuickTextControlPrivate()
       hasImState(false),
       cursorRectangleChanged(false),
       hoveredMarker(false),
+      selectByTouchDrag(false),
+      imSelectionAfterPress(false),
       lastSelectionStart(-1),
       lastSelectionEnd(-1)
 {}
@@ -323,9 +290,11 @@ void QQuickTextControlPrivate::setContent(Qt::TextFormat format, const QString &
             formatCursor.select(QTextCursor::Document);
             formatCursor.setCharFormat(charFormatForInsertion);
             formatCursor.endEditBlock();
+#if QT_CONFIG(textmarkdownreader)
         } else if (format == Qt::MarkdownText) {
             doc->setBaseUrl(doc->baseUrl().adjusted(QUrl::RemoveFilename));
             doc->setMarkdown(text);
+#endif
         } else {
 #if QT_CONFIG(texthtmlparser)
             doc->setHtml(text);
@@ -646,6 +615,12 @@ QTextDocument *QQuickTextControl::document() const
     return d->doc;
 }
 
+void QQuickTextControl::setDocument(QTextDocument *doc)
+{
+    Q_D(QQuickTextControl);
+    d->doc = doc;
+}
+
 void QQuickTextControl::updateCursorRectangle(bool force)
 {
     Q_D(QQuickTextControl);
@@ -774,8 +749,7 @@ void QQuickTextControl::processEvent(QEvent *e, const QTransform &transform)
         case QEvent::ShortcutOverride:
             if (d->interactionFlags & Qt::TextEditable) {
                 QKeyEvent* ke = static_cast<QKeyEvent *>(e);
-                if (isCommonTextEditShortcut(ke))
-                    ke->accept();
+                ke->setAccepted(isCommonTextEditShortcut(ke));
             }
             break;
         default:
@@ -945,6 +919,12 @@ void QQuickTextControlPrivate::keyPressEvent(QKeyEvent *e)
 process:
     {
         if (q->isAcceptableInput(e)) {
+#if QT_CONFIG(im)
+            // QTBUG-90362
+            // Before key press event will be handled, pre-editing part should be finished
+            if (isPreediting())
+                commitPreedit();
+#endif
             if (overwriteMode
                 // no need to call deleteChar() if we have a selection, insertText
                 // does it already
@@ -989,7 +969,7 @@ QRectF QQuickTextControlPrivate::rectForPosition(int position) const
         if (relativePos == preeditPos)
             relativePos += preeditCursor;
         else if (relativePos > preeditPos)
-            relativePos += layout->preeditAreaText().length();
+            relativePos += layout->preeditAreaText().size();
     }
 #endif
     QTextLine line = layout->lineForTextPosition(relativePos);
@@ -1019,6 +999,7 @@ void QQuickTextControlPrivate::mousePressEvent(QMouseEvent *e, const QPointF &po
 
     mousePressed = (interactionFlags & Qt::TextSelectableByMouse) && (e->button() & Qt::LeftButton);
     mousePressPos = pos.toPoint();
+    imSelectionAfterPress = false;
 
     if (sendMouseEventToInputContext(e, pos))
         return;
@@ -1032,6 +1013,9 @@ void QQuickTextControlPrivate::mousePressEvent(QMouseEvent *e, const QPointF &po
             cursor.clearSelection();
         }
     }
+
+    if (!selectByTouchDrag && !QQuickDeliveryAgentPrivate::isEventFromMouseOrTouchpad(e))
+        return;
     if (interactionFlags & Qt::TextEditable)
         blockWithMarkerUnderMousePress = q->blockWithMarkerAt(pos);
     if (e->button() & Qt::MiddleButton) {
@@ -1102,6 +1086,9 @@ void QQuickTextControlPrivate::mousePressEvent(QMouseEvent *e, const QPointF &po
 
 void QQuickTextControlPrivate::mouseMoveEvent(QMouseEvent *e, const QPointF &mousePos)
 {
+    if (!selectByTouchDrag && !QQuickDeliveryAgentPrivate::isEventFromMouseOrTouchpad(e))
+        return;
+
     Q_Q(QQuickTextControl);
 
     if ((e->buttons() & Qt::LeftButton)) {
@@ -1182,6 +1169,7 @@ void QQuickTextControlPrivate::mouseReleaseEvent(QMouseEvent *e, const QPointF &
 
     const QTextCursor oldSelection = cursor;
     const int oldCursorPos = cursor.position();
+    const bool isMouse = QQuickDeliveryAgentPrivate::isEventFromMouseOrTouchpad(e);
 
     if (mousePressed) {
         mousePressed = false;
@@ -1197,6 +1185,8 @@ void QQuickTextControlPrivate::mouseReleaseEvent(QMouseEvent *e, const QPointF &
             q->insertFromMimeData(md);
 #endif
     }
+    if (!isMouse && !selectByTouchDrag && !imSelectionAfterPress && interactionFlags.testFlag(Qt::TextEditable))
+        setCursorPosition(pos);
 
     repaintOldAndNewSelection(oldSelection);
 
@@ -1205,7 +1195,8 @@ void QQuickTextControlPrivate::mouseReleaseEvent(QMouseEvent *e, const QPointF &
         q->updateCursorRectangle(true);
     }
 
-    if ((interactionFlags & Qt::TextEditable) && (e->button() & Qt::LeftButton) && blockWithMarkerUnderMousePress.isValid()) {
+    if ((isMouse || selectByTouchDrag) && interactionFlags.testFlag(Qt::TextEditable) &&
+            (e->button() & Qt::LeftButton) && blockWithMarkerUnderMousePress.isValid()) {
         QTextBlock block = q->blockWithMarkerAt(pos);
         if (block == blockWithMarkerUnderMousePress) {
             auto fmt = block.blockFormat();
@@ -1243,7 +1234,8 @@ void QQuickTextControlPrivate::mouseDoubleClickEvent(QMouseEvent *e, const QPoin
 {
     Q_Q(QQuickTextControl);
 
-    if (e->button() == Qt::LeftButton && (interactionFlags & Qt::TextSelectableByMouse)) {
+    if (e->button() == Qt::LeftButton && (interactionFlags & Qt::TextSelectableByMouse)
+            && (selectByTouchDrag || QQuickDeliveryAgentPrivate::isEventFromMouseOrTouchpad(e))) {
 #if QT_CONFIG(im)
         commitPreedit();
 #endif
@@ -1287,7 +1279,7 @@ bool QQuickTextControlPrivate::sendMouseEventToInputContext(QMouseEvent *e, cons
         QTextLayout *layout = cursor.block().layout();
         int cursorPos = q->hitTest(pos, Qt::FuzzyHit) - cursor.position();
 
-        if (cursorPos >= 0 && cursorPos <= layout->preeditAreaText().length()) {
+        if (cursorPos >= 0 && cursorPos <= layout->preeditAreaText().size()) {
             if (e->type() == QEvent::MouseButtonRelease) {
                 QGuiApplication::inputMethod()->invokeAction(QInputMethod::Click, cursorPos);
             }
@@ -1306,10 +1298,11 @@ bool QQuickTextControlPrivate::sendMouseEventToInputContext(QMouseEvent *e, cons
 void QQuickTextControlPrivate::inputMethodEvent(QInputMethodEvent *e)
 {
     Q_Q(QQuickTextControl);
-    if (!(interactionFlags & Qt::TextEditable) || cursor.isNull()) {
+    if (cursor.isNull()) {
         e->ignore();
         return;
     }
+    bool textEditable = interactionFlags.testFlag(Qt::TextEditable);
     bool isGettingInput = !e->commitString().isEmpty()
             || e->preeditString() != cursor.block().layout()->preeditAreaText()
             || e->replacementLength() > 0;
@@ -1317,14 +1310,14 @@ void QQuickTextControlPrivate::inputMethodEvent(QInputMethodEvent *e)
     int oldCursorPos = cursor.position();
 
     cursor.beginEditBlock();
-    if (isGettingInput) {
+    if (isGettingInput && textEditable) {
         cursor.removeSelectedText();
     }
 
     QTextBlock block;
 
     // insert commit string
-    if (!e->commitString().isEmpty() || e->replacementLength()) {
+    if (textEditable && (!e->commitString().isEmpty() || e->replacementLength())) {
         if (e->commitString().endsWith(QChar::LineFeed))
             block = cursor.block(); // Remember the block where the preedit text is
         QTextCursor c = cursor;
@@ -1333,50 +1326,56 @@ void QQuickTextControlPrivate::inputMethodEvent(QInputMethodEvent *e)
         c.insertText(e->commitString());
     }
 
-    for (int i = 0; i < e->attributes().size(); ++i) {
-        const QInputMethodEvent::Attribute &a = e->attributes().at(i);
-        if (a.type == QInputMethodEvent::Selection) {
-            QTextCursor oldCursor = cursor;
-            int blockStart = a.start + cursor.block().position();
-            cursor.setPosition(blockStart, QTextCursor::MoveAnchor);
-            cursor.setPosition(blockStart + a.length, QTextCursor::KeepAnchor);
-            repaintOldAndNewSelection(oldCursor);
-            forceSelectionChanged = true;
+    if (interactionFlags & (Qt::TextSelectableByKeyboard | Qt::TextSelectableByMouse)) {
+        for (int i = 0; i < e->attributes().size(); ++i) {
+            const QInputMethodEvent::Attribute &a = e->attributes().at(i);
+            if (a.type == QInputMethodEvent::Selection) {
+                if (mousePressed)
+                    imSelectionAfterPress = true;
+                QTextCursor oldCursor = cursor;
+                int blockStart = a.start + cursor.block().position();
+                cursor.setPosition(blockStart, QTextCursor::MoveAnchor);
+                cursor.setPosition(blockStart + a.length, QTextCursor::KeepAnchor);
+                repaintOldAndNewSelection(oldCursor);
+                forceSelectionChanged = true;
+            }
         }
     }
 
     if (!block.isValid())
         block = cursor.block();
 
-    QTextLayout *layout = block.layout();
-    if (isGettingInput) {
-        layout->setPreeditArea(cursor.position() - block.position(), e->preeditString());
-        emit q->preeditTextChanged();
-    }
-    QVector<QTextLayout::FormatRange> overrides;
     const int oldPreeditCursor = preeditCursor;
-    preeditCursor = e->preeditString().length();
-    hasImState = !e->preeditString().isEmpty();
-    cursorVisible = true;
-    for (int i = 0; i < e->attributes().size(); ++i) {
-        const QInputMethodEvent::Attribute &a = e->attributes().at(i);
-        if (a.type == QInputMethodEvent::Cursor) {
-            hasImState = true;
-            preeditCursor = a.start;
-            cursorVisible = a.length != 0;
-        } else if (a.type == QInputMethodEvent::TextFormat) {
-            hasImState = true;
-            QTextCharFormat f = qvariant_cast<QTextFormat>(a.value).toCharFormat();
-            if (f.isValid()) {
-                QTextLayout::FormatRange o;
-                o.start = a.start + cursor.position() - block.position();
-                o.length = a.length;
-                o.format = f;
-                overrides.append(o);
+    if (textEditable) {
+        QTextLayout *layout = block.layout();
+        if (isGettingInput) {
+            layout->setPreeditArea(cursor.position() - block.position(), e->preeditString());
+            emit q->preeditTextChanged();
+        }
+        QVector<QTextLayout::FormatRange> overrides;
+        preeditCursor = e->preeditString().size();
+        hasImState = !e->preeditString().isEmpty();
+        cursorVisible = true;
+        for (int i = 0; i < e->attributes().size(); ++i) {
+            const QInputMethodEvent::Attribute &a = e->attributes().at(i);
+            if (a.type == QInputMethodEvent::Cursor) {
+                hasImState = true;
+                preeditCursor = a.start;
+                cursorVisible = a.length != 0;
+            } else if (a.type == QInputMethodEvent::TextFormat) {
+                hasImState = true;
+                QTextCharFormat f = qvariant_cast<QTextFormat>(a.value).toCharFormat();
+                if (f.isValid()) {
+                    QTextLayout::FormatRange o;
+                    o.start = a.start + cursor.position() - block.position();
+                    o.length = a.length;
+                    o.format = f;
+                    overrides.append(o);
+                }
             }
         }
+        layout->setFormats(overrides);
     }
-    layout->setFormats(overrides);
 
     cursor.endEditBlock();
 
@@ -1427,7 +1426,7 @@ QVariant QQuickTextControl::inputMethodQuery(Qt::InputMethodQuery property, cons
         QTextCursor tmpCursor = d->cursor;
         int localPos = d->cursor.position() - block.position();
         QString result = block.text().mid(localPos);
-        while (result.length() < maxLength) {
+        while (result.size() < maxLength) {
             int currentBlock = tmpCursor.blockNumber();
             tmpCursor.movePosition(QTextCursor::NextBlock);
             if (tmpCursor.blockNumber() == currentBlock)
@@ -1460,6 +1459,8 @@ QVariant QQuickTextControl::inputMethodQuery(Qt::InputMethodQuery property, cons
         result += QStringView{block.text()}.mid(0,localPos);
         return QVariant(result);
     }
+    case Qt::ImReadOnly:
+        return QVariant(!d->interactionFlags.testFlag(Qt::TextEditable));
     default:
         return QVariant();
     }
@@ -1630,6 +1631,14 @@ void QQuickTextControl::setWordSelectionEnabled(bool enabled)
     d->wordSelectionEnabled = enabled;
 }
 
+#if QT_VERSION < QT_VERSION_CHECK(7, 0, 0)
+void QQuickTextControl::setTouchDragSelectionEnabled(bool enabled)
+{
+    Q_D(QQuickTextControl);
+    d->selectByTouchDrag = enabled;
+}
+#endif
+
 QMimeData *QQuickTextControl::createMimeDataFromSelection() const
 {
     Q_D(const QQuickTextControl);
