@@ -2,21 +2,19 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 15.2.3.8-2-4
-description: Object.seal - non-enumerable own property of 'O' is sealed
+es5id: 15.2.3.8-2-a-8
+description: >
+    Object.seal - 'P' is own property of an Array object that uses
+    Object's [[GetOwnProperty]]
 includes: [propertyHelper.js]
 ---*/
 
-var obj = {};
+var obj = [];
 
-Object.defineProperty(obj, "foo", {
-  value: 10,
-  enumerable: false,
-  configurable: true
-});
+obj.foo = 10;
 
 assert(Object.isExtensible(obj));
 Object.seal(obj);
 
-assert(obj.hasOwnProperty("foo"));
 verifyNotConfigurable(obj, "foo");
+assert.sameValue(obj.foo, 10);
